unit heightmap;

{$mode objfpc}{$H+}
{$modeswitch nestedprocvars}

interface

uses
  Classes, SysUtils, ExtCtrls, Graph2D, multidim, mathconfig, math, Graphics
  {$IFDEF UNIX}, UTF8Process{$ENDIF};

type

  {
   TRenderer
   Multithreading for Hightmap rendering.
  }
  TRenderer = class (TThread)
   type
    TOperation = procedure(y1,y2:Integer) is nested;
   private
     suspendEv : PRTLEvent;
     y1, y2 : integer;
     busy : boolean;
     operation : TOperation;
   protected
    procedure execute; override;
   public
    constructor create;
    destructor destroy; override;
  end;

  {
   THeightMap
   Draws a function R²->R into a graph using a colored hightmap.
  }
  THeightMap = class(TGraph2D.TRenderObj)
   type
    SMD = specialize Multidimensional<MD2d, MD1d>;
    TBufferPoint = packed record
      x, y, z : Precision;
    end;
    PBufferPoint = ^TBufferPoint;
    TBuffer = Array of TBufferPoint;
   private
    _fkt : SMD.IFunction;
    buff : TBuffer;
    colorMin, colorMax : Precision;
    syncNext : THeightMap;
    class var threads : Array of TRenderer;
    class procedure writeMultithreading(v:Boolean); static;
    class function readMultithreading : Boolean; static;
    procedure sync(syncSrc:THeightMap);
    procedure writeFkt(v:SMD.IFunction);
   protected
    procedure render; override;
   public
    autoColorRange : Boolean;
    constructor create(_graph:TGraph2D); reintroduce;
    class property multithreading : boolean read readMultithreading write writeMultithreading;
    procedure synchronize(syncSrc:THeightMap);
    function synchronized : boolean;
    procedure desynchronize;
    property fkt : SMD.IFunction read _fkt write writeFkt;
    destructor destroy; override;
  end;

  {
   TGridPaint
   Draws grid points of a 2D grid into the graph.
  }
  TGridPaint = class (TGraph2D.TRenderObj)
   type
    SMD = specialize Multidimensional<MD2d, MD1d>;
    { TAccessableGrid }
    TAccessableGrid = object(SMD.TGrid)
      procedure setLane(dim:Word; lane:Cardinal; pos:Precision);
    end;
    PAccessableGrid = ^TAccessableGrid;
   strict private
    isMoving : Boolean;
    hoveredX, hoveredY : integer;
   protected
    procedure render; override;
    procedure mouseEvent(evType:TGraph2D.TMouseEventType; x,y:integer); override;
   public
    grid : SMD.PGrid;
    size : Byte;
    color : Cardinal;
    hoveredColor : Cardinal;
    editable : Boolean;
    onChange : TNotifyEvent;
    constructor create(_graph:TGraph2D);
  end;

implementation



{ TGridPaint.TAccessableGrid }

procedure TGridPaint.TAccessableGrid.setLane(dim: Word; lane: Cardinal; pos: Precision);
begin
  grid^[dim][lane] := pos;
end;



{ TGridPaint }

constructor TGridPaint.create(_graph:TGraph2D);
begin
  inherited create(_graph);
  grid := nil;
  color := $00FFFFFF;
  hoveredColor := $00FF00FF;
  hoveredX := -1;
  hoveredY := -1;
  size := 2;
  editable := false;
  isMoving := false;
  onChange := nil;
end;

procedure TGridPaint.render;
type
  TRow = array [0..MAXINT div 8] of Cardinal;
  PRow = ^TRow;
var
  i, j, k, l : integer;
  px, py : integer;
  sx, sy : Precision;
  row : PRow;
begin
  if (grid = nil) or (grid^.dimensions <> 2) then exit;
  graph.bmp.BeginUpdate;
  sx := (xAxis.screenToPos(1)-xAxis.screenToPos(0))*size/2;
  sy := (yAxis.screenToPos(1)-yAxis.screenToPos(0))*size/2;
  for i := 0 to grid^.laneCount[0]-1 do for j := 0 to grid^.laneCount[1]-1 do begin
    px := xAxis.posToScreen(grid^[0, i] - sx);
    py := yAxis.posToScreen(grid^[1, j] - sy);
    for k := px to px+size do for l := py to py+size do
      if (k >= 0) and (l >= 0) and (k < graph.bmp.width) and (l < graph.bmp.height) then begin
        row := graph.bmp.{%H-}scanLine[l];
        if (i = hoveredX) or (j = hoveredY) then row^[k] := hoveredColor
        else row^[k] := color;
      end;
    if (editable) then for k := px-3 to px+size+3 do for l := py-3 to py+size+3 do
      if (k >= 0) and (l >= 0) and (k < graph.bmp.width) and (l < graph.bmp.height) then begin
        graph.mouseMap[k, l] := self;
      end;
  end;
  if (isMoving) then for i := 0 to graph.width-1 do for j := 0 to graph.height-1 do graph.mouseMap[i, j] := self;
  graph.bmp.EndUpdate;
end;

procedure TGridPaint.mouseEvent(evType: TGraph2D.TMouseEventType; x, y: integer);
var
  p : Precision;
  b : boolean;
begin
  if (isMoving) then begin
    case (evType) of
     MouseUp: begin
      isMoving := false;
      graph.doPaint(self);
     end;
     MouseMove: begin
      p := xAxis.screenToPos(x);
      if ((hoveredX = 0) or (grid^[0, hoveredX-1] < p)) and ((hoveredX = grid^.laneCount[0]-1) or (grid^[0, hoveredX+1] > p)) then
        PAccessableGrid(grid)^.setLane(0, hoveredX, p);
      p := yAxis.screenToPos(y);
      if ((hoveredY = 0) or (grid^[1, hoveredY-1] < p)) and ((hoveredY = grid^.laneCount[1]-1) or (grid^[1, hoveredY+1] > p)) then
        PAccessableGrid(grid)^.setLane(1, hoveredY, p);
      if (assigned(onChange)) then onChange(self);
      graph.doPaint(self);
     end;
    end;
  end else begin
    case (evType) of
     MouseDown: begin
      if (hoveredX < 0) then exit;
      isMoving := true;
     end;
     MouseMove: begin
      if (hoveredX < 0) then b := true;
      p := xAxis.screenToPos(x);
      hoveredX := 0;
      while (hoveredX < grid^.laneCount[0]-1) and (abs(grid^[0, hoveredX+1] - p) < abs(grid^[0, hoveredX] - p)) do hoveredX += 1;
      p := yAxis.screenToPos(y);
      hoveredY := 0;
      while (hoveredY < grid^.laneCount[1]-1) and (abs(grid^[1, hoveredY+1] - p) < abs(grid^[1, hoveredY] - p)) do hoveredY += 1;
      if (b) then graph.doPaint(self);
     end;
     MouseOut: begin
      hoveredX := -1;
      hoveredY := -1;
      graph.doPaint(self);
     end;
    end;
  end;
end;




{ TRenderer }

constructor TRenderer.create;
begin
  freeOnTerminate := false;
  busy := true;
  suspendEv := RTLEventCreate;
  inherited create(false);
end;

procedure TRenderer.execute;
begin
  while (true) do begin
    busy := false;
    RTLEventWaitFor(suspendEv);
    if (terminated) then exit;
    operation(y1, y2);
  end;
end;

destructor TRenderer.destroy;
begin
  terminate;
  RtlEventSetEvent(suspendEv);
  inherited destroy;
  RTLeventdestroy(suspendEv);
end;




{ THeightMap }

constructor THeightMap.create(_graph:TGraph2D);
begin
  inherited create(_graph);
  buff := nil;
  fkt := nil;
  syncNext := self;
  autoColorRange := true;
end;

procedure THeightMap.synchronize(syncSrc: THeightMap);
var sn : THeightMap;
begin
  if (synchronized) then raise Exception.create('Object is already synchronized!');
  sn := syncSrc.syncNext; // store remaining loop until initial synchronisation is done. No need to let initial synchronisation circle.
  syncSrc.syncNext := self;
  self.syncNext := syncSrc;
  self.autoColorRange := false;
  sync(syncSrc);
  self.syncNext := sn;
end;

function THeightMap.synchronized: boolean;
begin
  result := syncNext <> self;
end;

procedure THeightMap.desynchronize;
var sn : THeightMap;
begin
  sn := self;
  repeat
    if (sn.syncNext = self) then begin
      sn.syncNext := self.syncNext;
      break;
    end else sn := sn.syncNext;
  until false;
  syncNext := self;
end;

class procedure THeightMap.writeMultithreading(v: Boolean);
var i : integer;
label wait;
begin
  if (v) then begin
    {$IFDEF UNIX}
    i := GetSystemThreadCount;
    {$ELSE}
    i := TThread.ProcessorCount;
    {$ENDIF}
    setLength(threads, i);
    for i := 0 to length(threads)-1 do threads[i] := TRenderer.create;
    wait:
    sleep(1);
    for i := 0 to length(threads)-1 do if (threads[i].busy) then goto wait;
  end else begin
    for i := 0 to length(threads)-1 do threads[i].destroy;
    setLength(threads, 0);
  end;
end;

class function THeightMap.readMultithreading: Boolean;
begin
  result := length(threads) <> 0;
end;

procedure THeightMap.sync(syncSrc: THeightMap);
var
  b : boolean;
begin
  if (syncSrc = self) then exit;
  b := false;
  if (colorMin <> syncSrc.colorMin) or (colorMax <> syncSrc.colorMax) then b := true;
  colorMin := syncSrc.colorMin;
  colorMax := syncSrc.colorMax;
  if (b) then graph.doPaint(self);
  syncNext.sync(syncSrc);
end;

procedure THeightMap.writeFkt(v: SMD.IFunction);
begin
  _fkt := v;
  if (length(buff) > 0) then fillChar(buff[0], length(buff)*sizeOf(TBufferPoint), #255);
end;


procedure THeightMap.render;
type
  TRow = array [0..MAXINT div 8] of Cardinal;
  PRow = ^TRow;
var
  accuracyDeltaX, accuracyDeltaY : Precision;
  ix, iy : integer;
  p : PBufferPoint;
  minMaxChanged : boolean;
  rows : Array of PRow;
label wait1, wait2, wait3;
const ColorMap : array [0..8] of Cardinal = ($FF000080,$FF3F48CC,$FF00A2E8,$FF22B14C,$FFB5E61D,$FFFFF200,$FFFF7F27,$FF880015,$FF3E2C26);

procedure calc(y1, y2 : integer);
var
  ix, iy : integer;
  x, y : Precision;
  p : PBufferPoint;
  p2 : MD2d;
begin
  for iy := y1 to y2 do for ix := 0 to graph.width-1 do begin // ensure, that buffer is correct
    x := xAxis.screenToPos(ix);
    y := yAxis.screenToPos(iy);
    p := @buff[iy*graph.width+ix];
    if (isNaN(p^.x)) or (abs(p^.x - x) > accuracyDeltaX) or (abs(p^.y - y) > accuracyDeltaY) or (graph.completeRepaint) then begin
      p^.x := x;
      p^.y := y;
      p2[0] := x;
      p2[1] := y;
      p^.z := fkt.calc(p2);
      minMaxChanged := true;
    end;
  end;
end;

procedure search(y1, y2 : integer);
var
  ix, iy : integer;
  p : PBufferPoint;
begin
  for iy := y1 to y2 do for ix := 0 to graph.width-1 do begin
    p := @buff[iy*graph.width+ix];
    if (p^.z < colorMin) then colorMin := p^.z;
    if (p^.z > colorMax) then colorMax := p^.z;
  end;
end;

procedure draw(y1, y2 : integer);
var
  ix, iy : integer;
  c1, c2 : cardinal;
  s : Precision;
begin
  for iy := y1 to y2 do for ix := 0 to graph.bmp.width-1 do begin
    s := (buff[graph.width*iy + ix].z-colorMin) / (colorMax-colorMin) * (length(ColorMap)-2);
    if (s < 0) then s := 0;
    if (s > Length(ColorMap)-1) then s := Length(ColorMap)-1;
    c1 := ColorMap[floor(s)];
    c2 := ColorMap[ceil(s)];
    s -= floor(s);
    rows[iy]^[ix] := $00000000
      or Trunc((c1 and $00FF0000)*(1-s)+(c2 and $00FF0000)*s) and $00FF0000
      or Trunc((c1 and $0000FF00)*(1-s)+(c2 and $0000FF00)*s) and $0000FF00
      or Trunc((c1 and $000000FF)*(1-s)+(c2 and $000000FF)*s) and $000000FF;
  end;
end;

begin
  if (fkt = nil) then exit;
  accuracyDeltaX := (xAxis.screenToPos(1)-xAxis.screenToPos(0)) / 8;
  accuracyDeltaY := (yAxis.screenToPos(1)-yAxis.screenToPos(0)) / 8;
  minMaxChanged := false;
  if (length(buff) < graph.width*graph.height) or (length(buff) > 2*graph.width*graph.height) then begin // ensure, that buffer size is correct
    ix := length(buff);
    setLength(buff, Trunc(graph.width*graph.height * 1.5));
    if (length(buff) > ix) then fillChar(buff[ix], (length(buff)-ix)*sizeOf(TBufferPoint), #255);
    minMaxChanged := true;
  end;
  if (multithreading) then begin
    for iy := 0 to length(threads)-1 do begin
      threads[iy].operation := @calc;
      threads[iy].y1 := round(graph.height / length(threads) * iy);
      threads[iy].y2 := round(graph.height / length(threads) * (iy+1))-1;
      threads[iy].busy := true;
      RtlEventSetEvent(threads[iy].suspendEv);
    end;
    wait1:
    sleep(1);
    for iy := 0 to length(threads)-1 do if (threads[iy].busy) then goto wait1;
  end else calc(0, graph.height-1);
  if (minMaxChanged) and (autoColorRange) then begin // compute min and max value for color table
    colorMin := Precision.MaxValue;
    colorMax := -Precision.MaxValue;
    if (multithreading) then begin
      for iy := 0 to length(threads)-1 do begin
        threads[iy].operation := @search;
        threads[iy].y1 := round(graph.bmp.height / length(threads) * iy);
        threads[iy].y2 := round(graph.bmp.height / length(threads) * (iy+1))-1;
        threads[iy].busy := true;
        RtlEventSetEvent(threads[iy].suspendEv);
      end;
      wait2:
      sleep(1);
      for iy := 0 to length(threads)-1 do if (threads[iy].busy) then goto wait2;
    end else search(0, graph.bmp.height-1);
    if (colorMin >= colorMax) then begin
      colorMin := colorMax-0.5;
      colorMax += 0.5;
    end;
    syncNext.sync(self);
  end;
  graph.bmp.beginUpdate;
  setLength(rows, graph.bmp.height);
  for iy := 0 to length(rows)-1 do rows[iy] := graph.bmp.{%H-}scanLine[iy];
  if (multithreading) then begin
    for iy := 0 to length(threads)-1 do begin
      threads[iy].operation := @draw;
      threads[iy].y1 := round(graph.bmp.height / length(threads) * iy);
      threads[iy].y2 := round(graph.bmp.height / length(threads) * (iy+1))-1;
      threads[iy].busy := true;
      RtlEventSetEvent(threads[iy].suspendEv);
    end;
    wait3:
    sleep(1);
    for iy := 0 to length(threads)-1 do if (threads[iy].busy) then goto wait3;
  end else draw(0, graph.bmp.height-1);
  graph.bmp.endUpdate;
end;

destructor THeightMap.destroy;
begin
  desynchronize;
  inherited destroy;
end;

initialization
setLength(THeightMap.threads, 0);

finalization
THeightMap.writeMultithreading(false);

end.

