unit linearalgebra;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, mathConfig;

type
  LinAlg = class
   type
    TVector = Array of Precision;

    { TMatrix }
    TMatrix = object
     type
      TMatrixInit = (INIT_NONE, INIT_NULL, INIT_IDENTITY);
     strict private
      buf    : Array of Precision;
      rows   : cardinal;
      colums : cardinal;
      function readEl(i,j:Cardinal) : Precision; inline;
      procedure writeEl(i,j:Cardinal; v:Precision); inline;
     public
      constructor init(nRows,nColums:Cardinal; initialize:TMatrixInit=INIT_NONE);
      property columCount : cardinal read colums;
      property rowCount : cardinal read rows;
      procedure switchRows(i1,i2:Cardinal);
      procedure switchColums(i1,i2:Cardinal);
      property el[i,j:Cardinal] : Precision read readEl write writeEl; default;
      function toString : String;
      destructor clear;
    end;

   public
    class procedure gauss(var M:TMatrix; out a:TVector; var b:TVector; completePivotSearch:Boolean=false);
  end;

implementation

{ LinAlg.TMatrix }

constructor LinAlg.TMatrix.init(nRows,nColums:Cardinal; initialize:TMatrixInit=INIT_NONE);
begin
  if (nRows = 0) or (nColums = 0) then raise Exception.Create('Empty Matrix!');
  rows := nRows;
  colums := nColums;
  setLength(buf, rows*colums);
  if (initialize <> INIT_NONE) then begin
    fillChar(buf[0], Length(buf)*sizeof(Precision), #0);
    if (initialize = INIT_IDENTITY) then
      while (nRows > 0) do begin
        nRows -= 1;
        el[nRows, nRows] := 1;
      end;
  end;
end;

function LinAlg.TMatrix.readEl(i, j: Cardinal): Precision;
begin
  {$IFOPT D+}
  if (i >= rows) or (j >= colums) then raise Exception.Create('Out of Bounds!');
  {$ENDIF}
  Result := buf[colums*i+j];
end;

procedure LinAlg.TMatrix.writeEl(i, j: Cardinal; v: Precision);
begin
  {$IFOPT D+}
  if (i >= rows) or (j >= colums) then raise Exception.Create('Out of Bounds!');
  {$ENDIF}
  buf[colums*i+j] := v;
end;

procedure LinAlg.TMatrix.switchRows(i1, i2: Cardinal);
var temp : Array of Precision;
begin
  {$IFOPT D+}
  if (i1 >= rows) or (i2 >= rows) then raise Exception.Create('Out of Bounds!');
  {$ENDIF}    setLength(temp, colums);
  move(buf[colums*i1], temp[0], colums*sizeOf(Precision));
  move(buf[colums*i2], buf[colums*i1], colums*sizeOf(Precision));
  move(temp[0], buf[colums*i2], colums*sizeOf(Precision));
end;

procedure LinAlg.TMatrix.switchColums(i1, i2: Cardinal);
var
  temp : Array of Precision;
  i : integer;
begin
  {$IFOPT D+}
  if (i1 >= colums) or (i2 >= colums) then raise Exception.Create('Out of Bounds!');
  {$ENDIF}
  setLength(temp, rows);
  for i := 0 to rows-1 do temp[i] := buf[colums*i + i1];
  for i := 0 to rows-1 do buf[colums*i + i1] := buf[colums*i + i2];
  for i := 0 to rows-1 do buf[colums*i + i2] := temp[i];
end;

function LinAlg.TMatrix.toString: String;
var i, j : integer;
begin
  result := '';
  for i := 0 to rows-1 do begin
    for j := 0 to colums-1 do
      result += FloatToStr(el[i, j], mathFloatFormat) + '     ';
    result += Char(13) + Char(10);
  end;
end;

destructor LinAlg.TMatrix.clear;
begin
  setLength(buf, 0);
  rows := 0;
  colums := 0;
end;



class procedure LinAlg.gauss(var M:TMatrix; out a:TVector; var b:TVector; completePivotSearch:Boolean=false);
var
  d, i, j, k, l, d_ : integer;
  p : Precision;
  aOrder : Array of cardinal;
begin
  if (M.rowCount <> M.columCount) then raise Exception.Create('Matrix is not quadratic!');
  if (Length(b) <> M.rowCount) then raise Exception.Create('Vector size does not fit matrix size!');
  if (completePivotSearch) then begin // umsortierung des Ergebnisvektors aufzeichnen
    setLength(aOrder, length(b));
    for i := 0 to length(aOrder)-1 do aOrder[i] := i;
  end;
  for d := 0 to M.rowCount-1 do begin // Obere Dreiecksmatrix erzeugen
    if (completePivotSearch) then for d_ := d to M.rowCount-2 do begin
      // Zeilen- und Spaltenpivotsuche - alle Zeilen unter d müssen sortiert werden, damit alle Nullen unten stehen
      i := d_; // i ist Zeilenindex mit maximalem Element
      j := d; // j ist Spaltenindex mit maximalem Element (nur suchen, wenn d_ = d)
      if (d_ = d) then begin
        for k := d_+1 to M.rowCount-1 do for l := d+1 to M.columCount-1 do
          if (abs(M[k, l]) > abs(M[i, j])) then begin i := k; j := l; end;
      end else
        for k := d_+1 to M.rowCount-1 do
          if (abs(M[k, d]) > abs(M[i, d])) then i := k;
      M.switchRows(d_, i);
      p := b[d_];
      b[d_] := b[i];
      b[i] := p;
      if (d = d_) then begin
        M.switchColums(d, j);
        k := aOrder[d];
        aOrder[d] := aOrder[j];
        aOrder[j] := k;
      end;
    end else for d_ := d to M.rowCount-2 do begin
      // Spaltenpivotsuche - alle Zeilen unter d müssen sortiert werden, damit alle Nullen unten stehen
      i := d_; // i ist Zeilenindex mit maximalem Element
      for j := d_+1 to M.rowCount-1 do
        if (abs(M[j, d]) > abs(M[i, d])) then i := j;
      M.switchRows(d_, i);
      p := b[d_];
      b[d_] := b[i];
      b[i] := p;
    end;
    if (M[d, d] = 0) then raise Exception.Create('System cannot be solved due to linear dependency at diagonal element '+inttostr(d)+'!');
    // Zeilen unter d,d Nullen
    for i := d+1 to M.rowCount-1 do begin
      if (M[i, d] = 0) then break;
      b[i] := b[i] - b[d] / M[d, d] * M[i, d];
      for j := M.columCount-1 downto d do
        M[i, j] := M[i, j] - M[d, j] / M[d, d] * M[i, d];
    end;
  end;
  setLength(a, Length(b));
  for d := M.columCount - 1 downto 0 do begin // Ergebnisvektor v berechnen
    p := 0;
    for i := d+1 to M.columCount-1 do p += a[i] * M[d, i];
    a[d] := (b[d] - p) / M[d, d];
  end;
  if (completePivotSearch) then begin // Ergebnisverktor zurück sortieren
    for i := 0 to length(aOrder)-2 do begin
      j := i; // index mit kleinstem Element ab i
      for k := i+1 to length(aOrder)-1 do
        if (aOrder[k] < aOrder[j]) then j := k;
      k := aOrder[i];
      aOrder[i] := aOrder[j];
      aOrder[j] := k;
      p := a[i];
      a[i] := a[j];
      a[j] := p;
    end;
  end;
end;

end.

