unit multidim;

{$mode objfpc}{$H+}
{$modeswitch nestedprocvars}
{$interfaces corba}

interface

uses
  Classes, SysUtils, mathConfig, math;

type
 MD1d = Precision;
 MD2d = array [0..1] of Precision;
 MD3d = array [0..2] of Precision;

 {
  MDnd
  This object contains a dynamic array of float and can be used to specialize
  Multidimensional for arbitrary dimensions.
 }
 MDnd = packed object
  private
   function readA(i: cardinal): Precision; inline;
   procedure writeA(i: cardinal; AValue: Precision); inline;
   function readCount : cardinal; inline;
   procedure writeCount(v:cardinal); inline;
  public
   a : array of Precision;
   property a_ [i:cardinal] : Precision read readA write writeA; default;
   property count : cardinal read readCount write writeCount;
  strict private
   {%H-}placeholder : byte; // this is to ensure, that sizeOf(MDnd) is not equal to any of the static array types!
 end;
 PMDnd = ^MDnd;

 {
  Multidimensional
  Contains and specializes some subclasses, which help to simplify the use of
  multidimensional math.
 }
 generic Multidimensional<Preimage{: MD1d or MD2d or MD3d or MDnd}, Image{: MD1d or MD2d or MD3d or MDnd}> = class
  type
   {
    TGridPoint
    Contains a static or dynamic array of cardinal, which has the length of
    Preimage (if static).
   }
   TGridPoint = packed object
    type
     TDynArray = Array of cardinal;
     PDynArray = ^TDynArray;
    strict private
     a : array [0..(sizeOf(Preimage) div sizeOf(Precision)) + (sizeOf(Preimage) and 1) * (sizeOf(TDynArray) div sizeOf(Cardinal) - sizeOf(Preimage) div sizeOf(Precision)) - 1] of cardinal; // if dynamic arrays are used, this array is large enough to store at least one array pointer
     function readCount : word; inline;
     procedure writeCount(c:word); inline;
     function readA(i:Word) : Cardinal; inline;
     procedure writeA(i:Word; v:Cardinal); inline;
    public
     constructor init(cnt:word);
     property count : word read readCount write writeCount;
     property arr[i:Word] : Cardinal read readA write writeA; default;
     destructor clear;
   end;
   TGridSize = TGridPoint;
   TAreaIterator = procedure (const p:Preimage) is nested;
   TGridIterator = procedure (const p:TGridPoint) of object;   // TODO: Replace the "of object" declaration with a nested method, as soon as FPC 3.1.1 is released

   {
    IFunction
    A general function Preimage --> Image.
   }
   IFunction = interface ['Multidimensional.IFunction']
     function calc(const v:Preimage) : Image;
     function toString : String;
     function asObject : TObject;
   end;

   {
    TGrid
    A multidimensional grid with non-equidistant steps (lanes).
    The lane positions are kept sorted to allow logarithmic search.
   }
   TGrid = object
    type
     TGridArray = Array [0..65535] of Array of Precision;
     PGridArray = ^TGridArray;
     TDynArray = Array of Array of Precision;
     PDynArray = ^TDynArray;
    strict private
     grid_static : array [0..(sizeOf(Preimage) div sizeOf(Precision) - 1)*(1-(sizeOf(Preimage) and 1))] of array of Precision; // if Preimage is MDnd, this array is still one element large ans stores the dynamic array of array of precision
     function readLane(dimension: word; laneIndex: cardinal): precision; inline;
     function readLaneCount(dimension:word) : cardinal; inline;
     function readGrid : PGridArray; inline;
    protected
     _onAddLane : procedure of object;
     property grid : PGridArray read readGrid;
     function mdIndexTo1DIndex(p: TGridPoint; isCellIndex:Boolean): integer; // helpfull to map the grid onto a 1D-array.
    public
     constructor init(dim:integer=-1);
     function addGridLane(dimension:word; lane:precision) : boolean;
     function dimensions : word; inline;
     property laneCount[dimension:word] : cardinal read readLaneCount;
     property lanes[dimension:word; laneIndex:cardinal] : precision read readLane; default;
     procedure pointToNearestIndex(const p:Preimage; var ip:TGridPoint);
     procedure iterate(it:TGridIterator);
     function totalLaneCount : cardinal;
     function totalCellCount(keepOneCell:Boolean=true) : cardinal; // keepOneCell=true means, that the cell count of dimensions containing one lane is still 1
     function clone : TGrid; inline;
     destructor clear;
   end;
   PGrid = ^TGrid;

   {
    IInterpolation
    A function, which can interpolate an other function.
   }
   IInterpolation = interface(IFunction) ['Multidimensional.IInterpolation']
     procedure computeFor(orig:IFunction; const grid:TGrid);
   end;

   {
    TBuffer
    A multidimensional data array with non-equidistant steps
   }
   TBuffer = class(IFunction)
    strict private
     _grid : TGrid;
     buff : array of Image;
     function readBuff(p: TGridPoint): Image; inline;
     procedure writeBuff(p: TGridPoint; AValue: Image); inline;
     function asObject : TObject;
     function readGrid : PGrid;
    public
     constructor create;
     procedure setGrid(__grid:TGrid; initialize:Boolean=true; initVal:Image=0);
     function setPoint(p: Preimage; v: Image) : Boolean;
     property grid : PGrid read readGrid;
     property buffer[p:TGridPoint] : Image read readBuff write writeBuff; default;
     function calc(const v:Preimage) : Image;
     destructor destroy; override;
   end;

   TMDArray = array [0..65535] of Precision;
   PMDArray = ^TMDArray;

  protected
   // Methods for accessing (Pre)image as scalar or static or dynamic array type
   class function dim(const a:Preimage) : cardinal; inline;
   class procedure setDim(var a:Preimage; l:cardinal); inline;
   class function arr(var a:Preimage) : PMDArray; inline;
   class function arrRead(const a:Preimage; i:Cardinal) : Precision; inline;
   class function arrI(var a:Image) : PMDArray; inline;
   class function fromDynamicPreimage (dyn:MDnd) : Preimage; inline;
   class function toDynamicPreimage (stat:Preimage) : MDnd; inline;
  public
   class procedure iterateArea(areaMin,areaMax:Preimage; steps:TGridSize; it:TAreaIterator);  // you need {$modeswitch nestedprocvars} in the unit, that calls this function!
   class function spatVolume(areaMin,areaMax:Preimage) : Precision;
   class function spatToStr(areaMin,areaMax:Preimage) : String;
   class procedure strToSpat(s:String; out areaMin:Preimage; out areaMax:Preimage);
 end;



 {
  MDAsDynamic
  Helps to convert from static preimage multidimensional functions to dynamic
  preimage multidimensional functions. Specialize this class with the static
  type and the subclasses will be the equivalent dynamic type representation.
 }
 generic MDAsDynamic<Preimage, Image> = class(specialize Multidimensional<MDnd, Image>)
  type
   SMnD = specialize Multidimensional<MDnd, Image>;
   SMD = specialize Multidimensional<Preimage, Image>;

   {
    Converts a Multidimensional.IFunction with static arraytype preimage to
    a dynamic preimage function.
   }
   TFunction = class(SMnD.IFunction)
    private
     function asObject : TObject;
    protected
     fkt : SMD.IFunction;
    public
     property staticFkt : SMD.IFunction read fkt;
     constructor create(_fkt:SMD.IFunction);
     function calc(const v:MDnd) : Image;
   end;

   {
    Converts a Multidimensional.IInterpolation with static arraytype preimage to
    a dynamic preimage interpolation.
   }
   TInterpolation = class(TFunction, SMnD.IInterpolation)
    public
     constructor create(_fkt:SMD.IInterpolation);
     class function fromDynamicGrid(const grid:SMnD.TGrid) : SMD.TGrid; inline;
     procedure computeFor(orig:SMnD.IFunction; const grid:SMnD.TGrid);
   end;
 end;

 const NaN = 0/0;

implementation



{ MDnd }

function MDnd.readA(i: cardinal): Precision;
begin
  result := a[i];
end;

procedure MDnd.writeA(i: cardinal; AValue: Precision);
begin
  a[i] := AValue;
end;

function MDnd.readCount: cardinal;
begin
  result := length(a);
end;

procedure MDnd.writeCount(v: cardinal);
begin
  setLength(a, v);
end;




{ Multidimensional }

class function Multidimensional.dim(const a: Preimage): cardinal;
begin
  if (SizeOf(a) = SizeOf(MDnd)) then {%H-}result := length(PMDnd(@a)^.a)
  else {%H-}result := SizeOf(a) div sizeof(Precision);
end;

class procedure Multidimensional.setDim(var a: Preimage; l: cardinal);
begin
  if (SizeOf(a) = SizeOf(MDnd)) then {%H-}setLength(PMDnd(@a)^.a, l)
  else {%H-}if (SizeOf(a) div sizeof(Precision) <> l) then raise Exception.Create('Wrong Dimension!');
end;

class function Multidimensional.arr(var a:Preimage) : PMDArray;
begin
  if (SizeOf(a) = SizeOf(MDnd)) then {%H-}begin
   if (Length(PMDnd(@a)^.a) = 0) then result := nil
   else result := PMDArray(@PMDnd(@a)^.a[0]);
  end else {%H-}result := PMDArray(@a);
end;

class function Multidimensional.arrRead(const a:Preimage; i:Cardinal) : Precision;
begin
  if (SizeOf(a) = SizeOf(MDnd)) then {%H-}result := PMDnd(@a)^.a[i]
  else {%H-}result := PMDArray(@a)^[i];
end;

class function Multidimensional.arrI(var a:Image) : PMDArray;
begin
  if (SizeOf(a) = SizeOf(MDnd)) then {%H-}begin
   if (Length(PMDnd(@a)^.a) = 0) then result := nil
   else result := PMDArray(@PMDnd(@a)^.a[0]);
  end else {%H-}result := PMDArray(@a);
end;

class function Multidimensional.fromDynamicPreimage(dyn: MDnd): Preimage;
begin
  if (sizeOf(result) = sizeOf(MDnd)) then {%H-}begin
    PMDnd(@result)^.a := dyn.a;
    SetLength(PMDnd(@result)^.a, length(PMDnd(@result)^.a));
  end else {%H-}begin
    {$IFOPT D+}if (length(dyn.a) <> dim(result)) then raise Exception.Create('Wrong dimension!');{$ENDIF}
    move(dyn.a[0], result, length(dyn.a)*sizeOf(Precision));
  end;
end;

class function Multidimensional.toDynamicPreimage(stat: Preimage): MDnd;
begin
  if (sizeOf(stat) = sizeOf(MDnd)) then {%H-}begin
    result.a := PMDnd(@stat)^.a;
    SetLength(result.a, length(result.a));
  end else {%H-}begin
    SetLength(result.a, dim(stat));
    move(stat, result.a[0], dim(stat)*sizeOf(Precision));
  end;
end;

class procedure Multidimensional.iterateArea(areaMin, areaMax: Preimage; steps: TGridSize; it: TAreaIterator);
var
  p : Preimage;
  i : integer;
label repeatSearch;
begin
  {$IFOPT D+}if (dim(areaMin) <> dim(areaMax)) then raise Exception.create('Dimensions do not fit.');{$ENDIF}
  setDim(p{%H-}, dim(areaMin));
  move(arr(areaMin)^[0], arr(p)^[0], dim(p)*SizeOf(Precision));
  repeatSearch:
    it(p);
    for i := 0 to dim(p)-1 do begin
      arr(p)^[i] += (arr(areaMax)^[i]-arr(areaMin)^[i]) / steps[i];
      if (arr(p)^[i] <= arr(areaMax)^[i]) or (arr(p)^[i] < arr(areaMax)^[i] + (arr(areaMax)^[i]-arr(areaMin)^[i]) / (steps[i]*4)) then goto repeatSearch;
      arr(p)^[i] := arr(areaMin)^[i];
    end;
end;

class function Multidimensional.spatVolume(areaMin, areaMax: Preimage): Precision;
var d : integer;
begin
  {$IFOPT D+}if (dim(areaMin) <> dim(areaMax)) then raise Exception.create('Dimensions do not fit.');{$ENDIF}
  result := 1;
  for d := 0 to dim(areaMin)-1 do result *= arr(areaMax)^[d]-arr(areaMin)^[d];
end;

class function Multidimensional.spatToStr(areaMin, areaMax: Preimage): String;
var
  d : integer;
begin
  {$IFOPT D+}if (dim(areaMin) <> dim(areaMax)) then raise Exception.create('Dimensions do not fit.');{$ENDIF}
  result := '[';
  for d := 0 to dim(areaMin)-1 do begin
    if (d <> 0) then result += ', ';
    result += FloatToStr(arr(areaMin)^[d], mathFloatFormat) + '..' + FloatToStr(arr(areaMax)^[d], mathFloatFormat);
  end;
  result += ']';
end;

class procedure Multidimensional.strToSpat(s: String; out areaMin: Preimage; out areaMax: Preimage);
var
  i, ia : integer;
  state : byte;
  a : Array of Precision;
  str : ShortString;
begin
  str := '';
  state := 0;
  setLength(a, length(s) div 2);
  ia := 0;
  for i := 1 to length(s) do case (state) of
   0 : // search for spat starting bracket
    if (s[i] = '[') then state := 1;
   1 : // step over whitespaces until beginning of lower bound
    if (s[i] = ']') then state := 5
    else if (byte(s[i]) >= 48) and (byte(s[i]) < 58) or (s[i] = '-') then begin
     str := s[i];
     state := 2;
    end else if (s[i] <> #9) and (s[i] <> ' ') and (s[i] <> #10) and (s[i] <> #13) then
     raise Exception.Create('Unexpected char "'+s[i]+'" at position '+inttostr(i));
   2 : // read lower bound until ..
    if (s[i] = '.') and (i < length(s)) and (s[i+1] = '.') then begin
     a[ia] := StrToFloat(str, mathFloatFormat);
     ia += 1;
     state := 3;
    end else begin
     if (s[i] = ' ') then continue;
     if (s[i] = ',') then raise Exception.Create('Missing upper bound at position '+inttostr(i));
     if (length(str) >= 250) then raise Exception.Create('Too many chars for one number!');
     str += s[i];
    end;
   3 : // step over second .
    begin
     str := '';
     state := 4;
    end;
   4 : // read upper bound
    if (s[i] = ',') or (s[i] = ']') then begin
     a[ia] := StrToFloat(str, mathFloatFormat);
     ia += 1;
     if (s[i] = ']') then state := 5
     else state := 1;
    end else begin
     if (s[i] = ' ') or (s[i] = #9) or (s[i] = #13) then continue;
     if (length(str) >= 250) then raise Exception.Create('Too many chars for one number!');
     str += s[i];
    end;
   5 : // closing bracket was found
    break;
  end;
  if (state < 5) then raise Exception.create('Missing closing bracket!');
  if (ia and 1 <> 0) then raise Exception.create('Unknown error: Number of floats is uneven!');
  ia := ia div 2;
  setDim(areaMin{%H-}, ia);
  setDim(areaMax{%H-}, ia);
  for i := 0 to ia-1 do begin
    arr(areaMin)^[i] := a[i*2];
    arr(areaMax)^[i] := a[i*2+1];
  end;
end;




{ TGridPoint }

constructor Multidimensional.TGridPoint.init(cnt:word);
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}fillChar(a[0], sizeOf(a), #0);
  writeCount(cnt);
end;

function Multidimensional.TGridPoint.readCount : word; inline;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}result := length(PDynArray(@a[0])^)
  else {%H-}result := sizeOf(Preimage) div sizeOf(Precision);
end;

procedure Multidimensional.TGridPoint.writeCount(c:word); inline;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}setLength(PDynArray(@a[0])^, c)
  {$IFOPT D+}else {%H-}if (sizeOf(Preimage) div sizeOf(Precision) <> c) then raise Exception.create('Wrong Dimension!');{$ENDIF}
end;

function Multidimensional.TGridPoint.readA(i:Word) : Cardinal; inline;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}result := PDynArray(@a[0])^[i]
  else {%H-}result := a[i];
end;

procedure Multidimensional.TGridPoint.writeA(i:Word; v:Cardinal); inline;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}PDynArray(@a[0])^[i] := v
  else {%H-}a[i] := v;
end;

destructor Multidimensional.TGridPoint.clear;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}writeCount(0);
end;




{ TGrid }

constructor Multidimensional.TGrid.init(dim: integer = -1);
var d : integer;
begin
  if (sizeOf(Preimage) = sizeOf(MDnd)) then {%H-}begin
    if (dim = -1) then raise Exception.create('For dynamic multidimensional type, you have to specify the grid dimensions!');
    setLength(PDynArray(@grid_static[0])^, dim);
  end;// else setLength(grid, sizeOf(Preimage) div sizeOf(Precision));
  for d := 0 to dimensions-1 do grid^[d] := nil;
  _onAddLane := nil;
end;

function Multidimensional.TGrid.readGrid : PGridArray;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}result := @(PDynArray(@grid_static[0])^[0])
  else {%H-}result := @grid_static[0];
end;

function Multidimensional.TGrid.readLaneCount(dimension: word): cardinal;
begin
  {$IFOPT D+}if (dimension >= dimensions) then raise Exception.Create('Dimension greather than grid dimensions!');{$ENDIF}
  Result := length(grid^[dimension]);
end;

function Multidimensional.TGrid.readLane(dimension: word; laneIndex: cardinal): precision;
begin
  {$IFOPT D+}
  if (dimension >= dimensions) or (laneIndex >= length(grid^[dimension])) then raise Exception.Create('Index out of bounds!');
  {$ENDIF}
  Result := grid^[dimension][laneIndex];
end;

function Multidimensional.TGrid.addGridLane(dimension: word; lane: precision) : boolean;
var i, j : integer;
begin
  result := false;
  {$IFOPT D+}if (dimension >= dimensions) then raise Exception.Create('Dimension greather than grid dimensions!');{$ENDIF}
  i := length(grid^[dimension])-1;
  while (i >= 0) do
    if (grid^[dimension][i] > lane) then i -= 1
    else if (grid^[dimension][i] = lane) then exit // lane already exists
    else break;
  setLength(grid^[dimension], length(grid^[dimension])+1);
  for j := length(grid^[dimension])-1 downto i+2 do grid^[dimension][j] := grid^[dimension][j-1];
  grid^[dimension][i+1] := lane;
  result := true;
  if (assigned(_onAddLane)) then _onAddLane;
end;

function Multidimensional.TGrid.dimensions: word;
begin
  if (sizeOf(Preimage) and 1 = 1) then {%H-}result := length(PDynArray(@grid_static[0])^)
  else {%H-}result := sizeOf(Preimage) div sizeOf(Precision);
end;

procedure Multidimensional.TGrid.pointToNearestIndex(const p: Preimage; var ip: TGridPoint);
var i, min, max, m : integer;
begin
  {$IFOPT D+}if (ip.count <> dim(p)) then raise Exception.Create('Lengths do not match!');{$ENDIF}
  for i := 0 to dim(p)-1 do begin
    min := 0;
    max := laneCount[i]-1;
    if (max = -1) then raise Exception.Create('One grid dimension is empty!');
    while (min < max-1) do begin
      m := (min+max) div 2;
      if (grid^[i][m] > arrRead(p, i)) then max := m
      else min := m;
    end;
    if (arrRead(p, i) >= (grid^[i][min]+grid^[i][max]) / 2) then ip[i] := max
    else ip[i] := min;
  end;
end;

procedure Multidimensional.TGrid.iterate(it: TGridIterator);
var
  i : integer;
  p : TGridPoint;
begin
  {%H-}p.count := dimensions;
  for i := 0 to p.count-1 do begin
    if (laneCount[i] = 0) then exit;
    p[i] := 0;
  end;
  i := 0;
  while (i < p.count) do begin
    if (p[i] >= laneCount[i]) then begin
      p[i] := 0;
      i += 1;
    end else begin
      it(p);
      p[i] := p[i] + 1;
    end;
  end;
end;

function Multidimensional.TGrid.mdIndexTo1DIndex(p: TGridPoint; isCellIndex:Boolean): integer;
var
  dimSize : cardinal;
  i, l : integer;
begin
  {$IFOPT D+}if (p.count <> dimensions) then raise Exception.Create('Wrong dimension!');{$ENDIF}
  dimSize := 1;
  result := 0;
  for i := 0 to p.count-1 do begin
    l := laneCount[i];
    if (isCellIndex) and (l <> 1) then l -= 1;
    if (p[i] >= l) then begin
      result := -1;
      exit;
    end;
    result += p[i]*dimSize;
    dimSize *= l;
  end;
end;

function Multidimensional.TGrid.totalLaneCount: cardinal;
var d : integer;
begin
  result := 0;
  for d := dimensions-1 downto 0 do result += laneCount[d];
end;

function Multidimensional.TGrid.totalCellCount(keepOneCell:Boolean=true): cardinal;
var d : integer;
begin
  result := 1;
  for d := dimensions-1 downto 0 do
    if (laneCount[d] > 1) then result *= laneCount[d] - 1
    else if (laneCount[d] = 0) or (not keepOneCell) then result := 0;
end;

function Multidimensional.TGrid.clone: TGrid;
var i : integer;
begin
  result.init(dimensions);
  for i := 0 to dimensions-1 do begin
    setLength(result.grid^[i], length(grid^[i]));
    if (length(grid^[i]) > 0) then move(grid^[i][0], result.grid^[i][0], length(grid^[i])*sizeOf(Precision));
  end;
end;

destructor Multidimensional.TGrid.clear;
var i : integer;
begin
  for i := 0 to dimensions-1 do setLength(grid^[i], 0);
  if (sizeOf(Preimage) and 1 = 1) then {%H-}setLength(PDynArray(@grid_static[0])^, 0);
end;




{ TMDBuffer }

constructor Multidimensional.TBuffer.create;
begin
  inherited;
  _grid.init;
  buff := nil;
end;

function Multidimensional.TBuffer.readBuff(p: TGridPoint): Image;
begin
  Result := buff[_grid.mdIndexTo1DIndex(p, false)];
end;

procedure Multidimensional.TBuffer.writeBuff(p: TGridPoint; AValue: Image);
begin
  buff[_grid.mdIndexTo1DIndex(p, false)] := AValue;
end;

procedure Multidimensional.TBuffer.setGrid(__grid: TGrid; initialize:Boolean=true; initVal:Image=0);
var i,j : integer;
begin
  _grid.clear;
  _grid := __grid.clone;
  j := 1;
  for i := 0 to _grid.dimensions-1 do j *= _grid.laneCount[i];
  setLength(buff, j);
  if (initialize) then for i := 0 to length(buff)-1 do buff[i] := initVal;
  if (j = 0) then raise Exception.create('One grid dimension is empty!');
end;

function Multidimensional.TBuffer.setPoint(p: Preimage; v: Image) : Boolean;
var
  i, j, k : integer;
  newBuff : Array of Image;
  newI, addI : TGridPoint;
label
  noCopyOld;
begin
  {$IFOPT D+}if (dim(p) <> _grid.dimensions) then raise Exception.create('Wrong dimension!');{$ENDIF}
  {%H-}addI.count := dim(p);
  for i := 0 to dim(p)-1 do if (_grid.addGridLane(i, arr(p)^[i])) then begin // add new lanes, if neccessary
    for j := _grid.laneCount[i]-1 downto 0 do if (_grid[i, j] <= arr(p)^[i]) then begin  // find the index of the new lanes and store in addI
      addI[i] := j;
      break;
    end;
  end else addI[i] := Cardinal.maxValue;
  j := 1;
  for i := 0 to _grid.dimensions-1 do j *= _grid.laneCount[i];
  if (j <> length(buff)) then begin // at least one lane was added, buffer has to be copied
    setLength(newBuff, j);
    {%H-}newI.count := dim(p);
    for i := 0 to newI.count-1 do newI[i] := 0;
    k := 0;
    for i := 0 to length(newBuff)-1 do begin
      for j := 0 to dim(p)-1 do if (newI[j] = addI[j]) then begin
        arrI(newBuff[i])^[0] := 0/0; // init new buffer points with NaN
        goto noCopyOld;
      end;
      newBuff[i] := buff[k]; // copy old buffer points
      k += 1;
      noCopyOld:
      for j := 0 to dim(p)-1 do begin  // inc newI;
        newI[j] := newI[j] + 1;
        if (newI[j] >= _grid.laneCount[j]) then newI[j] := 0
        else break;
      end;
    end;
    buff := newBuff;
  end;
  for i := 0 to dim(p)-1 do if (addI[i] = Cardinal.maxValue) then  // find all missing indices
    for j := 0 to _grid.laneCount[i]-1 do if (_grid[i, j] <= arr(p)^[i]) then begin
      addI[i] := j;
      break;
    end;
  i := _grid.mdIndexTo1DIndex(addI, false);
  result := isNaN(arrI(buff[i])^[0]);
  buff[i] := v;
end;

function Multidimensional.TBuffer.calc(const v: Preimage): Image;
var
  p : TGridPoint;
  i : integer;
  min, max, m : integer;
begin
  p.init(dim(v));
  for i := 0 to p.count-1 do begin
    min := 0;
    max := _grid.laneCount[i]-1;
    if (max < 0) then begin
      fillChar(result, sizeOf(result), #0);
      exit;
    end;
    while (min < max-1) do begin
      m := (min+max) div 2;
      if (_grid[i, m] > arrRead(v, i)) then max := m
      else min := m;
    end;
    if (arrRead(v, i) >= (_grid[i, min]+_grid[i, max])/2) then p[i] := max
    else p[i] := min;
  end;
  result := readBuff(p);
end;

function Multidimensional.TBuffer.asObject: TObject;
begin
  Result := self;
end;

function Multidimensional.TBuffer.readGrid: PGrid;
begin
  result := @_grid;
end;

destructor Multidimensional.TBuffer.destroy;
begin
  setLength(buff, 0);
  _grid.clear;
  inherited;
end;




{ MDAsDynamic.TFunction }

constructor MDAsDynamic.TFunction.create(_fkt: SMD.IFunction);
begin
  inherited create;
  fkt := _fkt;
end;

function MDAsDynamic.TFunction.asObject: TObject;
begin
  result := self;
end;

function MDAsDynamic.TFunction.calc(const v: MDnd): Image;
begin
  result := fkt.calc(SMD.fromDynamicPreimage(v));
end;




{ MDAsDynamic.TInterpolation }

constructor MDAsDynamic.TInterpolation.create(_fkt:SMD.IInterpolation);
begin
  inherited create(_fkt);
end;

class function MDAsDynamic.TInterpolation.fromDynamicGrid(const grid:SMnD.TGrid) : SMD.TGrid; inline;
var i : integer;
begin
  result.init(grid.dimensions);
  for i := 0 to result.dimensions-1 do begin
    result.grid^[i] := grid.grid^[i];
  end;
end;

procedure MDAsDynamic.TInterpolation.computeFor(orig:SMnD.IFunction; const grid:SMnD.TGrid);
begin
  if (not (orig.asObject is TFunction)) then raise Exception.create('Orig must be of type MDAsDynamic.TFunction!');
  (fkt.asObject as SMD.IInterpolation).computeFor(TFunction(orig.asObject).fkt, fromDynamicGrid(grid));
end;


initialization
if (sizeOf(Precision) and 1 <> 0) then {%H-}raise Exception.Create('Precision must not have an uneven size!');

end.


