unit numericAnalysis;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, mathconfig, math, multidim;

type
  {$interfaces corba}
  Analysis = class
   type
    TMD11 = specialize Multidimensional<MD1d, MD1d>;
    {
     IAnalyticFunction
     General function, which can be derived or intergrated.
    }
    IFunction = interface(TMD11.IFunction) ['Analysis.IFunction']
      function derivate : IFunction;
      function integrate(c:Precision) : IFunction;
    end;

    {
     TNumericFunction
     Provides several numerical implementations of analytic operations.
    }
    TNumericFunction = class (IFunction)
     private
      function asObject : TObject;
     public
      function calc(const x : Precision) : Precision; virtual; abstract;
      function derivate : Analysis.IFunction; virtual; abstract;
      function integrate(c : Precision) : IFunction; virtual; abstract;
      function trapezIntegral(min, max : Precision; n : Cardinal) : Precision;
      function newton(y : Precision; df : IFunction; var x : Precision; min : Precision = 0; max : Precision = 0; gradient : Precision = 0) : Boolean;
      function calcInverse(y, a, b : Precision; df : IFunction) : Precision;
    end;
  end;

implementation

procedure destroyInterface (var inst: Analysis.IFunction);
var o : TObject;
begin
  o := inst.asObject;
  inst := nil;
  o.destroy;
end;

{ Numeric.TNumericFunction }

function Analysis.TNumericFunction.trapezIntegral(min, max: Precision; n: Cardinal): Precision;
var
  h : Precision;
  i : integer;
  A : Precision;
begin
  Result := 0;
  h := (max - min) / n;
  for i := 1 to n do begin
    A := (Calc(min + (i-1) * h) + Calc(min + i*h)) * 0.5 * h;
    if (not IsNaN(A)) then Result := Result + A;
  end;
end;

function Analysis.TNumericFunction.newton(y: Precision; df: IFunction; var x: Precision; min: Precision; max: Precision; gradient: Precision): Boolean;
var
  s, d, f, f_, f0 : Precision;
  n : Byte;
begin
  Result := false;
  f := self.Calc(x) - y;
  n := 0;
  f0 := f*f;
  f_ := df.Calc(x);
  if (f = 0) and (f_ * Gradient <= 0) and (Gradient <> 0) then exit;
  f_ := 0;
  while (n < SizeOf(Precision) div 16 + 5) and (f <> 0) do begin
    s := df.Calc(x);
    if (s = 0) or (s * gradient < 0) then if (f_ = 0) and (min <> max) then begin
      if (x > (max + min)/2) then x := x - (max-min)/2
      else x := x + (max-min)/2;
      f_ := 1;
      continue;
    end else exit;
    s := - f / s;
    d := 2;   // d ist Amijo-Sigma, wird in erstem Schleifendurchlauf halbiert!
    repeat begin
      d := d / 2;
      if (d*s = 0) then exit;
      if ((x + d*s < min) or (x + d*s > max)) and (max <> min) then begin f_ := 2*f; continue; end;
      if (gradient <> 0) and (df.Calc(x + d*s) * gradient < 0) then begin f_ := 2*f; continue; end;
      f_ := self.Calc(x + d*s) - y;
    end until (f_*f_ <= f*f*(1-d*0.001));
    if (f = f_) then break;
    f := f_;
    x := x + d*s;
    if (n <> 0) then n := n + 1
    else if (f*f < f0/10) then n := 1;
  end;
  Result := f*f < 1E-20;
end;

function Analysis.TNumericFunction.calcInverse(y, a, b: Precision; df: IFunction): Precision;
var c : Boolean;
begin
  // Ungelöstes Problem: Globalisiertes Newtonverfahren kann aus [a, b] herausspringen, auch wenn es eine Nullstelle in [a, b] gibt!
  if (df = nil) then begin c := true; df := self.Derivate; end else c := false;
  Result := (a + b) / 2;
  try
    self.Newton(y, df, Result);
    if (Result < a) or (Result > b) then Result := NaN;
  except
    Result := NaN;
  end;
  if (c) then destroyInterface(df);
end;

function Analysis.TNumericFunction.asObject: TObject;
begin
  Result := self;
end;



end.

