unit splines;

{$mode objfpc}{$H+}
{$modeswitch nestedprocvars}

interface

uses
  Classes, SysUtils, multidim, linearalgebra, mathConfig, numericAnalysis, poly,
  math;

{$IFNDEF NOCOMPILE}{$DEFINE COMPILE}{$ENDIF} // this is to get rid of errors, that the code completion has with generic and subclasses

type
  Spline = class
   type

    {
     TGeneral
     A general function composed of peacewise defined "spline" functions.
    }
    {$IFDEF COMPILE}generic{$ENDIF} TGeneral{$IFDEF COMPILE}<Preimage, Image>{$ENDIF} = class (TObject, specialize Multidimensional<Preimage, Image>{$IFDEF COMPILE}.IFunction{$ENDIF})
     type
      SMD = specialize Multidimensional<Preimage, Image>;
      { TGrid }
      TGrid = object (SMD.TGrid)
       type
        procedureOfObject = procedure of object;
       public
        procedure setLane(l:integer; val:Precision);
        function index1D(p:SMD.TGridPoint; isCell:Boolean) : integer;
        property onAddLane : procedureOfObject write _onAddLane;
      end;
     strict private
      appliedGrid : TGrid;
      gridApplied : Boolean;
      function readSpline(p:SMD.TGridPoint) : SMD.IFunction;
      procedure writeSpline(p: SMD.TGridPoint; AValue: SMD.IFunction);
      procedure gridLaneAdded;
      function asObject : TObject;
     protected
      spl : Array of SMD.IFunction;
      procedure applyGrid;
     public
      periodic : array of boolean;
      grid : TGrid;
      constructor create; reintroduce;
      function toString : String; override;
      function calc(const v:Preimage) : Image;
      property splines[p:SMD.TGridPoint] : SMD.IFunction read readSpline write writeSpline;
      destructor destroy; override;
    end;

    {$IFDEF COMPILE}generic{$ENDIF} TLinear{$IFDEF COMPILE}<Preimage, Image>{$ENDIF} = class (specialize Spline.TGeneral<Preimage, Image>, specialize Multidimensional<Preimage, Precision>{$IFDEF COMPILE}.IInterpolation{$ENDIF})
     type
      TLinSpl = class (SMD.IFunction)
       private
        co : Array of Precision; // TODO: replace with static array [0..(2**(sizeOf(Preimage) div sizeOf(MD1d)))-1]
        function asObject : TObject;
       public
        function calc(const v:Preimage) : Image;
      end;
     public
      procedure computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid);
    end;

    TGeneral1D = specialize TGeneral<MD1d, MD1d>;

    {
     TAnalytical
     A spline function R->R, which can be derived or integrated.
    }
    TAnalytical = class (TGeneral1D, Analysis.IFunction)
     type
      SG = specialize TGeneral<MD1d, MD1d>;
     public
      function derivate : Analysis.IFunction;
      function integrate(c:Precision) : Analysis.IFunction;
    end;

    {
     TC2Spline
     Represents a cubic interpolation of a 1D grid. The interpolation points
     were taken from an existing IMDFunction.
    }

    { TCubicC2 }

    TCubicC2 = class(TAnalytical, specialize Multidimensional<MD1d, MD1d>{$IFDEF COMPILE}.IInterpolation{$ENDIF})
     type
      TBordaryCondition = (NATURAL, HERMITE, PERIODICAL, NOT_A_KNOT);
     strict private
      function readSpline(i:cardinal) : TPolynom;
     public
      procedure computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid); overload;
      procedure computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid; bc:TBordaryCondition; fa:Precision=0; fb:Precision=0); overload;
      property splines[i:cardinal] : TPolynom read readSpline;
    end;

    {
     TBicubicC2
     Represents a bicubic interpolation of a 2D grid. The interpolation points
     were taken from an existing IMDFunction
    }
    TBicubicC2 = class(specialize TGeneral<MD2d, MD1d>, specialize Multidimensional<MD2d, MD1d>{$IFDEF COMPILE}.IInterpolation{$ENDIF})
     type
      SMD1 = specialize Multidimensional<MD1d, MD1d>;
      SMD2 = specialize Multidimensional<MD2d, MD1d>;
      {
       TSplPoly
       A polynom of degree 3 in two dimensions
      }
      TSplPoly = class(SMD2.IFunction)
       private
        co : Array [0..3, 0..3] of Precision;
        function asObject : TObject;
       public
        function calc(const v:MD2d) : Precision;
        function derivative(dx,dy:Byte; x,y:Precision) : Precision;
      end;
      {
       TBufferCut
       Represents a cut through a 2D function as a 1D function.
       Used to compute sx and sy for the bicubic interpolation.
      }
      TBufferCut = class(SMD1.IFunction)
       private
        function asObject : TObject;
       public
        buffer : SMD2.IFunction;
        cutPos : Precision;
        cutDir : byte;
        function calc(const v:Precision) : Precision;
      end;
      {
       TSplineDerivative
       Returns the derivation of the spline, which is closest to x at the
       value of splinePos. Used to compute the sd for the bicubic interpolation.
      }
      TSplineDerivative = class(SMD1.IFunction)
       private
        function asObject : TObject;
       public
        splines : Array of TCubicC2;
        grid : SMD1.PGrid;
        splinePos : Precision;
        function calc(const v:Precision) : Precision; overload;
      end;
     private
      function asObject : TObject;
      function readSpline(p:SMD2.TGridPoint) : TSplPoly;
     public
      bordaryCondition : TCubicC2.TBordaryCondition;
      constructor create;
      property splines[p:SMD2.TGridPoint] : TSplPoly read readSpline;
      function calc(const v:MD2d) : Precision;
      function derivative(dx,dy:Byte; x,y:Precision) : Precision;
      procedure computeFor(orig:SMD2.IFunction; const _grid:SMD2.TGrid);
    end;

  end;


implementation



procedure Spline.TGeneral.TGrid.setLane(l: integer; val: Precision);
begin
  grid^[0][l] := val;
end;

function Spline.TGeneral.TGrid.index1D(p:SMD.TGridPoint; isCell:Boolean) : integer;
begin
  result := mdIndexTo1DIndex(p, isCell);
end;



{ TGeneralSpline }

constructor Spline.TGeneral.create;
begin
  inherited;
  spl := nil;
  periodic := nil;
  grid.init;
  appliedGrid.init;
  grid.onAddLane := @gridLaneAdded;
  gridApplied := true;
end;

procedure Spline.TGeneral.gridLaneAdded;
begin
  gridApplied := false;
end;

procedure Spline.TGeneral.applyGrid;
var
  i, j : integer;
begin
  i := grid.totalCellCount;
  for j := i to length(spl)-1 do if (spl[j] <> nil) then spl[j].asObject.destroy;
  j := length(spl);
  setLength(spl, i);
  for j := j to length(spl)-1 do spl[j] := nil;
  appliedGrid := TGrid(grid.clone);
  gridApplied := true;
end;

function Spline.TGeneral.readSpline(p: SMD.TGridPoint): SMD.IFunction;
begin
  if (not gridApplied) then applyGrid;
  Result := spl[grid.mdIndexTo1DIndex(p, true)];
end;

procedure Spline.TGeneral.writeSpline(p: SMD.TGridPoint; AValue: SMD.IFunction);
var
  i : integer;
  o : TObject;
begin
  if (not gridApplied) then applyGrid;
  i := grid.mdIndexTo1DIndex(p, true);
  if (spl[i] <> nil) then begin
    o := spl[i].asObject;
    spl[i] := nil;
    o.destroy;
  end;
  spl[i] := AValue;
end;

function Spline.TGeneral.calc(const v: Preimage): Image;
var
  p : SMD.TGridPoint;
  min, max, m, i : integer;
  vp : Preimage;
begin
  if (not gridApplied) then applyGrid;
  p.init(SMD.dim(v));
  vp := v;
  for i := 0 to p.count-1 do begin
    if (i < length(periodic)) and (periodic[i]) then begin // make periodic
      SMD.setDim(vp, SMD.dim(vp));
      m := floor((SMD.arrRead(vp, i) - grid[i, 0]) / (grid[i, grid.laneCount[i]-1] - grid[i, 0]));
      SMD.arr(vp)^[i] := SMD.arrRead(vp, i) - m*(grid[i, grid.laneCount[i]-1] - grid[i, 0]);
    end;
    min := 0;
    max := grid.laneCount[i]-1;
    if (max < 0) then begin result := 0; exit; end;
    while (min < max-1) do begin
      m := (min+max) div 2;
      if (grid[i, m] > SMD.arrRead(vp, i)) then max := m
      else min := m;
    end;
    p[i] := min;
  end;
  i := grid.mdIndexTo1DIndex(p, true);
  if (i < 0) or (spl[i] = nil) then result := 0
  else result := spl[i].calc(vp);
  p.clear;
end;

function Spline.TGeneral.toString : String;
var
  p : SMD.TGridPoint;
  d, i : integer;
label
  nextDimIncrement;
begin
  Result := '';
  if (grid.dimensions = 0) then exit;
  p.init(grid.dimensions);
  while (length(spl) > 0) do begin
    i := grid.mdIndexTo1DIndex(p, true);
    result += '[';
    for d := 0 to grid.dimensions-1 do begin
      if (p[d] > 0) then result += floatToStr(grid.lanes[d, p[d]])+' < ';
      result += 'x';
      if (grid.dimensions > 1) then result += intToStr(d);
      if (p[d] < grid.laneCount[d]-2) then result += ' ≤ '+floatToStr(grid.lanes[d, p[d]+1])
      else if (p[d] = 0) then result += ' ∈ ℝ';
    end;
    result += ': '+spl[i].toString+']';
    d := grid.dimensions-1;
    nextDimIncrement: begin
      p[d] := p[d] + 1;
      if (p[d] >= grid.laneCount[d]-1) then begin
        p[d] := 0;
        d -= 1;
        if (d < 0) then break
        else goto nextDimIncrement;
      end;
    end;
  end;
end;

function Spline.TGeneral.asObject: TObject;
begin
  result := self;
end;

destructor Spline.TGeneral.destroy;
var
  i : integer;
  o : TObject;
begin
  for i := 0 to length(spl)-1 do if (spl[i] <> nil) then begin
    o := spl[i].asObject;
    spl[i] := nil;
    o.destroy;
  end;
  grid.clear;
  appliedGrid.clear;
  inherited;
end;



{ TAnalytic }

function Spline.TAnalytical.derivate : Analysis.IFunction;
var
  r : TAnalytical;
  i : integer;
begin
  r := TAnalytical.create;
  result := r;
  setLength(r.spl, length(spl));
  r.grid := SG.TGrid(grid.clone);
  for i := 0 to length(spl)-1 do r.spl[i] := (spl[i].asObject as Analysis.IFunction).derivate;
end;

function Spline.TAnalytical.integrate(c:Precision) : Analysis.IFunction;
var
  r : TAnalytical;
  i : integer;
begin
  r := TAnalytical.create;
  Result := r;
  setLength(r.spl, length(spl));
  r.grid := SG.TGrid(grid.clone);
  for i := 0 to length(spl)-1 do r.spl[i] := (spl[i].asObject as Analysis.IFunction).integrate(c);
end;




{ TLinear }

function Spline.TLinear.TLinSpl.asObject : TObject;
begin
  result := self;
end;

function Spline.TLinear.TLinSpl.calc(const v:Preimage) : Image;
var
  w : word;
  i : integer;
  s : Precision;
begin
  result := 0;
  for i := 0 to length(co)-1 do begin
    s := co[i];
    for w := 0 to SMD.dim(v)-1 do if (i and (1 shl w) <> 0) then s *= SMD.arrRead(v, w);
    result += s;
  end;
end;

procedure Spline.TLinear.computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid);
var
  i, j : integer;
  s : Precision;
  w : word;
  M : LinAlg.TMatrix;
  a, b : LinAlg.TVector;
  v : Preimage;
  p : SMD.TGridPoint;
  coLen : cardinal;
begin
  coLen := 2**_grid.dimensions;
  // apply grid
  grid.clear;
  grid := TGrid(_grid.clone);
  applyGrid;
  for i := 0 to length(spl)-1 do if (spl[i] = nil) then begin
    spl[i] := Spline.TLinear.TLinSpl.create as SMD.IFunction;
    setLength(TLinSpl(spl[i].asObject).co, coLen);
  end;
  for i := 0 to _grid.dimensions-1 do if (grid.laneCount[i] = 0) then exit; // abort, if one dimension is null
  // compute spline parts
  p.init(_grid.dimensions);
  for i := 0 to p.count-1 do p[i] := 0;
  M.init(coLen, coLen);
  setLength(b, coLen);
  while (p[p.count-1] < _grid.laneCount[p.count-1]-1) or (_grid.laneCount[p.count-1] = 1) and (p[p.count-1] = 0) do begin // iterate through all grid cells
    for i := 0 to coLen-1 do begin // iterate M's and b's rows
      for j := 0 to coLen-1 do begin // iterate M's colums
        s := 1;
        for w := 0 to _grid.dimensions-1 do // built product of dimensional parts (1 | x | y | x*y | ...)
          if (j and (1 shl w) <> 0) then
            if (_grid.laneCount[w] > 1) then s *= _grid[w, p[w] + ((i shr w) and 1)]
            else s *= _grid[w, 0] + ((i shr w) and 1) * (1 + _grid[w, 0]); // special case: only one lant in dimension w
        M[i, j] := s;
      end;
      SMD.setDim(v{%H-}, _grid.dimensions);
      for w := 0 to _grid.dimensions-1 do // select edge point corresponding to row i
        if (_grid.laneCount[w] > 1) then SMD.arr(v)^[w] := _grid[w, p[w] + ((i shr w) and 1)]
        else SMD.arr(v)^[w] := _grid[w, 0]; // special case: only one lane in dimension w
      b[i] := orig.calc(v); // calc orig value of selected edge point
    end;
    LinAlg.gauss(M, a, b, false); // calc coefficients
    move(a[0], TLinSpl(splines[p].asObject).co[0], coLen * sizeOf(Precision)); // apply coefficients
    for i := 0 to p.count-1 do begin // increment cell index
      p[i] := p[i] + 1;
      if (p[i] < _grid.laneCount[i]-1) then break
      else if (i < p.count-1) then p[i] := 0;
    end;
  end;
  M.clear;
  setLength(b, 0);
  p.clear;
end;




{ TCubicC2 }

function Spline.TCubicC2.readSpline(i:cardinal) : TPolynom;
begin
  if (spl[i] = nil) then result := nil
  else result := TPolynom(spl[i].asObject);
end;

procedure Spline.TCubicC2.computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid);
begin
  computeFor(orig, _grid, NATURAL);
end;

procedure Spline.TCubicC2.computeFor(orig:SMD.IFunction; const _grid:SMD.TGrid; bc:TBordaryCondition; fa:Precision=0; fb:Precision=0);
type
  Row = record
    d : Precision;  // diagonal element
    dn : Precision; // next element behind d (except of the last row)
    a : Precision;  // left hand side result vector - also stores the last Element of a matrix row in case of periodic bordary
    b : Precision;  // right hand side vector
    y : Precision;  // buffered result of orig
  end;
var
  m        : Array of Row;
  c,d,f    : precision;
  j,n      : integer;
  specCase : byte;
procedure specialCaseSpline; // this procedure computes the not-a-knot spline for lower boundary in the n-a-k special case
var
  inf : TPolynom.TInterpolateInfo;
  i,j,n : integer;
  p : TPolynom;
begin
  if (grid.laneCount[0] <= 4) then begin // this is the common case for not-a-knot for up to 4 fixed points
    setLength(inf, grid.laneCount[0]);
    for i := 0 to length(inf)-1 do begin
      inf[i].x := grid[0, i];
      inf[i].y := orig.calc(inf[i].x);
      inf[i].d := 0;
    end;
    n := length(inf)-2;
  end else begin // this is the special case for equidistant points at lower bordary
    setLength(inf, 4);
    inf[0].x := grid[0,1];
    inf[0].y := m[1].y;
    inf[0].d := 0;
    inf[1].x := grid[0,2];
    inf[1].y := m[2].y;
    inf[1].d := 0;
    inf[2].x := grid[0,2];
    inf[2].y := 3*splines[2].co[3]*grid[0,2]**2 + 2*splines[2].co[2]*grid[0,2] + splines[2].co[1];
    inf[2].d := 1;
    inf[3].x := grid[0,2];
    inf[3].y := 6*splines[2].co[3]*grid[0,2] + 2*splines[2].co[2];
    inf[3].d := 2;
    n := 1;
  end;
  p := TPolynom.hermiteInterpolation(inf);
  for i := 0 to n do
    for j := 0 to 3 do
      if (j <= p.degree) then splines[i].co[j] := p.co[j]
      else splines[i].co[j] := 0;
  p.destroy;
end;
begin
  if (_grid.dimensions <> 1) then raise Exception.Create('Only one dimension is allowed!');
  grid.clear;
  setLength(periodic, 0);
  grid := TGrid(_grid.clone);
  applyGrid;
  for j := 0 to length(spl)-1 do if (spl[j] = nil) then spl[j] := (TPolynom.create(3) as Analysis.IFunction);
  if (grid.laneCount[0] <= 4) then begin
    if (grid.laneCount[0] = 0) then exit
    else if (grid.laneCount[0] = 1) then begin
      splines[0].co[0] := orig.calc(grid[0, 0]);
      splines[0].co[1] := 0;
      splines[0].co[2] := 0;
      splines[0].co[3] := 0;
      exit;
    end else if (bc = NOT_A_KNOT) then begin
      specialCaseSpline; // for four or less points, not-a-knot is a simple polynom, but the usual computation here can fail
      exit;
    end;
  end;
  // Lower bordary condition
  n := grid.laneCount[0]-1; // n is the number of intervals between the fixed points
  specCase := 0;
  setLength(m, n+1); // set the number of equations to solve
  m[0].y := orig.calc(grid[0, 0]);
  m[1].y := orig.calc(grid[0, 1]);
  j := 1;
  case (bc) of
   NATURAL : begin
     m[0].d  := 1;
     m[0].dn := 0;
     m[0].b  := 0;

     // y-1 = 0
     m[0].d := (grid[0, 1] - grid[0, 0]) * (2/3);
     m[0].dn := (grid[0, 1] - grid[0, 0])/6 * (8/7);
     m[0].b := (m[1].y - 9/4 * m[0].y)/(grid[0, 1] - grid[0, 0]) * (8/7);

     // from wiki
     m[0].d := 2*(grid[0, 1] - grid[0, 0])/3;
     m[0].dn := (grid[0, 1] - grid[0, 0])/6;
     m[0].b := 3*m[1].y / (grid[0, 1] - grid[0, 0]);
   end;
   HERMITE : begin
     f := grid[0, 1] - grid[0, 0];
     m[0].d  := f / 3;
     m[0].dn := f / 6;
     m[0].b  := (m[1].y - m[0].y) / f - fa;
   end;
   PERIODICAL : begin
     setLength(periodic, 1);
     periodic[0] := true;
     n -= 1; // in periodical case, the last fixed point belongs to the next period
     m[0].d := (grid[0, 1] - grid[0, 0] + grid[0, n+1] - grid[0, n]) / 3;
     m[0].dn := (grid[0, 1] - grid[0, 0]) / 6;
     m[0].a := (grid[0, n+1] - grid[0, n]) / 6; // upper right edge element
     m[0].b := (m[1].y - m[0].y) / (grid[0, 1] - grid[0, 0])  -  (m[0].y - orig.calc(grid[0, n])) / (grid[0, n+1] - grid[0, n]);
   end;
   NOT_A_KNOT : begin
     f := (grid[0,2]-grid[0,1])-(grid[0,1]-grid[0,0]);
     m[2].y := orig.calc(grid[0, 2]);
     if (f = 0) then begin // special case: not-a-knot condition for equidistant points leads to division by zero; requires limes computation up to m[2]
       specCase := 2; // mark this as special case
       j := 3; // shift begin of common matrix computation
       f := (m[2].y - 2*m[1].y + m[0].y) / (grid[0,1] - grid[0,0]);
       m[3].y := orig.calc(grid[0, 3]);
       m[2].d := (grid[0,3] - grid[0,1]) / 3;
       m[2].dn := (grid[0, 3] - grid[0, 2]) / 6;
       m[2].b := (m[3].y - m[2].y) / (grid[0, 3] - grid[0, 2])  -  (m[2].y - m[1].y) / (grid[0, 2] - grid[0, 1])  -  f/6
     end else begin
       m[0].d := 1;
       m[0].dn := -((grid[0,2]-grid[0,1])+2*(grid[0,1]-grid[0,0]))/f;
       m[0].b := -6*((m[2].y - m[1].y)/(grid[0,2]-grid[0,1]) - (m[1].y - m[0].y)/(grid[0,1]-grid[0,0]))*(grid[0,1]-grid[0,0])/((grid[0,2]-grid[0,1])**2-(grid[0,1]-grid[0,0])**2);
     end;
   end;
  end;
  // Store upper triangular matrix in m (all entries below diagonal are already zero - this is the subtraction with f*[...])
  for j := j to n-1 do begin
    f := (grid[0, j] - grid[0, j-1]) / (6*m[j-1].d);
    m[j+1].y := orig.calc(grid[0, j+1]);
    m[j].d := (grid[0, j+1] - grid[0, j-1]) / 3  -  f * m[j-1].dn;
    m[j].dn := (grid[0, j+1] - grid[0, j]) / 6;
    m[j].b := (m[j+1].y - m[j].y) / (grid[0, j+1] - grid[0, j])  -  (m[j].y - m[j-1].y) / (grid[0, j] - grid[0, j-1])  -  f * m[j-1].b;
    if (bc = PERIODICAL) then m[j].a := 0  -  f*m[j-1].a;
  end;
  // upper bordary condition
  case (bc) of
   NATURAL : begin
     m[n].d  := 1;
     m[n].dn := 0;
     m[n].b  := 0;
   end;
   HERMITE : begin
     f := (grid[0, n] - grid[0, n-1]) / (6 * m[n-1].d);
     m[n].d := (grid[0, n] - grid[0, n-1]) / 3  -  f * m[n-1].dn;
     m[n].dn := 0;
     m[n].b := fb  -  (m[n].y - m[n-1].y) / (grid[0, n] - grid[0, n-1])  -  f * m[n-1].b;
   end;
   PERIODICAL : begin
     if (n = 0) then raise Exception.create('For functions with two points, periodical boundary fails, as second point is just used to determine the interval!');
     m[n].d := (grid[0, n] - grid[0, n-1] + grid[0, n+1] - grid[0, n]) / 3;
     m[n].dn := 0;
     m[n].b := (m[0].y - m[n].y) / (grid[0, n+1] - grid[0, n])  -  (m[n].y - m[n-1].y) / (grid[0, n] - grid[0, n-1]);
     f := (grid[0, n+1] - grid[0, n]) / 6; // lower left edge element
     for j := 0 to n-2 do begin
       f := f / m[j].d;
       m[n].d := m[n].d  -  f * m[j].a;
       m[n].b := m[n].b  -  f * m[j].b;
       f := 0  -  f * m[j].dn;
     end;
     f += (grid[0, n] - grid[0, n-1]) / 6; // add the value of lambda_n
     f := f / m[n-1].d;
     m[n].d := m[n].d  -  f * (m[n-1].dn+m[n-1].a);
     m[n].b := m[n].b  -  f * m[n-1].b;
   end;
   NOT_A_KNOT : begin
     f := (grid[0, n-1] - grid[0, n-2]) / (grid[0, n] - grid[0, n-1]);
     m[n].d := 1;
     m[n].dn := 0;
     m[n].b := (m[n-1].b/m[n-1].d*(1+f) - m[n-2].b/m[n-2].d + m[n-1].b*m[n-2].dn/(m[n-1].d*m[n-2].d))  /  (m[n-1].dn*m[n-2].dn/(m[n-1].d*m[n-2].d) + m[n-1].dn/m[n-1].d*(1+f) + f);
   end;
  end;
  // solve m / compute a-vector
  m[n].a := m[n].b / m[n].d;
  if (bc = PERIODICAL) then begin
    for j := n-1 downto specCase do m[j].a := (m[j].b - m[j+1].a*m[j].dn - m[n].a*m[j].a) / m[j].d;
    m[n+1].y := m[0].y;
    m[n+1].a := m[0].a;
    n += 1;
  end else for j := n-1 downto specCase do m[j].a := (m[j].b - m[j+1].a*m[j].dn) / m[j].d;
  // compose polynoms
  for j := specCase to n-1 do begin
    f := grid[0, j+1] - grid[0, j];
    c := (m[j+1].y - m[j].y) / f  -  (m[j+1].a - m[j].a) * f / 6;
    d := m[j].y - m[j].a * f*f/6;
    splines[j].co[3] := (m[j].a                   - m[j+1].a                ) / (-6*f);
    splines[j].co[2] := (m[j].a * grid[0, j+1]    - m[j+1].a * grid[0, j]   ) / ( 2*f);
    splines[j].co[1] := (m[j].a * grid[0, j+1]**2 - m[j+1].a * grid[0, j]**2) / (-2*f) + c;
    splines[j].co[0] := (m[j].a * grid[0, j+1]**3 - m[j+1].a * grid[0, j]**3) / ( 6*f) - c*grid[0, j] + d;
  end;
  if (specCase <> 0) then specialCaseSpline;
end;




{ TSplPoly }

function Spline.TBicubicC2.TSplPoly.asObject : TObject;
begin
  result := self;
end;

function Spline.TBicubicC2.TSplPoly.calc(const v:MD2d): Precision;
var x2, x3, y_ : Precision;
begin
  x2 := v[0]*v[0];
  x3 := x2*v[0];
  Result := co[0,0]      + co[1,0]*v[0]      + co[2,0]*x2      + co[3,0]*x3;
  Result += co[0,1]*v[1] + co[1,1]*v[0]*v[1] + co[2,1]*x2*v[1] + co[3,1]*x3*v[1];
  y_ := v[1]*v[1];
  Result += co[0,2]*y_   + co[1,2]*v[0]*y_   + co[2,2]*x2*y_   + co[3,2]*x3*y_;
  y_ *= v[1];
  Result += co[0,3]*y_   + co[1,3]*v[0]*y_   + co[2,3]*x2*y_   + co[3,3]*x3*y_;
end;

function Spline.TBicubicC2.TSplPoly.derivative(dx, dy: Byte; x, y: Precision): Precision;
var ix, iy : integer;
procedure summand;
var
  a : Precision;
  j : integer;
begin
  a := co[ix,iy];
  j := ix;
  while (j > ix-dx) do begin
    a *= j;
    j -= 1;
  end;
  while (j > 0) do begin
    a *= x;
    j -= 1;
  end;
  j := iy;
  while (j > iy-dy) do begin
    a *= j;
    j -= 1;
  end;
  while (j > 0) do begin
    a *= y;
    j -= 1;
  end;
  result += a;
end;
begin
  result := 0;
  for ix := 0 to 3 do for iy := 0 to 3 do summand;
end;


{ TSplineDerivative }

function Spline.TBicubicC2.TSplineDerivative.asObject : TObject;
begin
  result := self;
end;

function Spline.TBicubicC2.TSplineDerivative.calc(const v: Precision): Precision;
var m, max, i, j : integer;
function derivativeAt(p:TPolynom; x:Precision) : precision; inline;
begin
  Result := 3*p.co[3]*x*x + 2*p.co[2]*x + p.co[1];
end;
begin
  if (grid^.laneCount[0] = 0) then begin
    Result := 0;
    exit;
  end;
  // search i - the index in the splines-array
  i := 0;
  max := grid^.laneCount[0]-1;
  while (i < max) do begin
    m := (i+max+1) div 2;
    if (grid^[0, m] > v) then max := m-1
    else i := m;
  end;
  // search j - the index of the correct polynom in splines[i]
  j := 0;
  max := splines[i].grid.totalCellCount-1;
  while (j < max) do begin
    m := (j+max+1) div 2;
    if (splines[i].grid[0, m] > splinePos) then max := m-1
    else j := m;
  end;
  // calc
  result := derivativeAt(splines[i].splines[j], splinePos);
end;


{ TBufferCut }

function Spline.TBicubicC2.TBufferCut.asObject : TObject;
begin
  result := self;
end;

function Spline.TBicubicC2.TBufferCut.calc(const v: Precision): Precision;
var v2 : Array [0..1] of Precision;
begin
  if (cutDir = 0) then begin
    v2[1] := cutPos;
    v2[0] := v;
  end else begin
    v2[0] := cutPos;
    v2[1] := v;
  end;
  result := buffer.calc(v2);
end;


{ TBicubicC2 }

function Spline.TBicubicC2.asObject : TObject;
begin
  result := self;
end;

function Spline.TBicubicC2.readSpline(p:SMD2.TGridPoint) : TSplPoly;
begin
  result := TsplPoly(spl[grid.mdIndexTo1DIndex(p, true)].asObject);
end;

constructor Spline.TBicubicC2.create;
begin
  inherited;
  bordaryCondition := NATURAL;
end;

function Spline.TBicubicC2.calc(const v:MD2d): Precision;
begin
  result := derivative(0, 0, v[0], v[1]);
end;

function Spline.TBicubicC2.derivative(dx, dy: Byte; x, y: Precision): Precision;
var
  xi, yi : integer;
  m, max : integer;
begin
  if (grid.laneCount[0] = 0) or (grid.laneCount[1] = 0) then begin
    Result := 0;
    exit;
  end;
  if (length(periodic) = 2) then begin // make periodic
    if (periodic[0]) then begin
      m := floor((x - grid[0, 0]) / (grid[0, grid.laneCount[0]-1] - grid[0, 0]));
      x := x - m*(grid[0, grid.laneCount[0]-1] - grid[0, 0]);
    end;
    if (periodic[1]) then begin
      m := floor((y - grid[1, 0]) / (grid[1, grid.laneCount[1]-1] - grid[1, 0]));
      y := y - m*(grid[1, grid.laneCount[1]-1] - grid[1, 0]);
    end;
  end;
  // search xi
  xi := 0;
  max := grid.laneCount[0]-2;
  while (xi < max) do begin
    m := (xi+max+1) div 2;
    if (grid[0, m] > x) then max := m-1
    else xi := m;
  end;
  // search yi
  yi := 0;
  max := grid.laneCount[1]-2;
  while (yi < max) do begin
    m := (yi+max+1) div 2;
    if (grid[1, m] > y) then max := m-1
    else yi := m;
  end;
  // calc
  result := TSplPoly(spl[yi*(grid.laneCount[0]-1) + xi].asObject).derivative(dx, dy, x, y);
end;

procedure Spline.TBicubicC2.computeFor(orig: SMD2.IFunction; const _grid: SMD2.TGrid);
var
  xi, yi : integer;
  p : array [0..1] of Precision;
  sy, sx, sd : Array of TCubicC2;
  cut : TBufferCut;
  der : TSplineDerivative;
  cutGrid : SMD1.TGrid;
  M : LinAlg.TMatrix;
  a, b : LinAlg.TVector;
  x1, x1_2, x1_3, x2, x2_2, x2_3, y1, y1_2, y1_3, y2, y2_2, y2_3 : Precision;
procedure Mrow (row : cardinal; m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, m11, m12, m13, m14, m15 : Precision); inline;
begin
  M[row, 0] := m0; M[row, 1] := m1; M[row, 2] := m2; M[row, 3] := m3; M[row, 4] := m4; M[row, 5] := m5; M[row, 6] := m6; M[row, 7] := m7; M[row, 8] := m8; M[row, 9] := m9; M[row, 10] := m10; M[row, 11] := m11; M[row, 12] := m12; M[row, 13] := m13; M[row, 14] := m14; M[row, 15] := m15;
end;
begin
  if (_grid.dimensions <> 2) then raise Exception.Create('Only two dimensions are allowed!');
  grid.clear;
  grid := TGrid(_grid.clone);
  applyGrid;
  for xi := 0 to length(spl)-1 do if (spl[xi] = nil) then spl[xi] := Spline.TBicubicC2.TSplPoly.create as SMD.IFunction;
  if (grid.laneCount[0] = 0) or (grid.laneCount[1] = 0) then exit;
  if (bordaryCondition = PERIODICAL) then begin
    setLength(periodic, 2);
    periodic[0] := true;
    periodic[1] := true;
  end else setLength(periodic, 0);
  // compute sy (splines in y-direction at several x values)
  cut := TBufferCut.create;
  cut.buffer := orig;
  cut.cutDir := 1;
  cutGrid.init;
  for yi := 0 to grid.laneCount[1]-1 do cutGrid.addGridLane(0, grid[1, yi]);
  setLength(sy, grid.laneCount[0]);
  for xi := 0 to grid.laneCount[0]-1 do begin
    cut.cutPos := grid[0, xi];
    sy[xi] := TCubicC2.create;
    sy[xi].computeFor(cut, cutGrid, bordaryCondition);
  end;
  // compute sx (splines in x-direction at several y values)
  cut.cutDir := 0;
  cutGrid.clear;
  cutGrid.init;
  for xi := 0 to grid.laneCount[0]-1 do cutGrid.addGridLane(0, grid[0, xi]);
  setLength(sx, grid.laneCount[1]);
  for yi := 0 to grid.laneCount[1]-1 do begin
    cut.cutPos := grid[1, yi];
    sx[yi] := TCubicC2.create;
    sx[yi].computeFor(cut, cutGrid, bordaryCondition);
  end;
  cut.destroy;
  // compute sd (splines in x-direction at several y values, interpolating the derivatives of the sy)
  der := TSplineDerivative.create;
  der.splines := sy;
  der.grid := @cutGrid;
  setLength(sd, grid.laneCount[1]);
  for yi := 0 to grid.laneCount[1]-1 do begin
    der.splinePos := grid[1, yi];
    sd[yi] := TCubicC2.create;
    sd[yi].computeFor(der, cutGrid, bordaryCondition);
  end;
  der.destroy;
  cutGrid.clear;
  // merge spline borders to 2D cubic flat
  try
    M.init(16, 16);
    setLength(b, 16);
    if (grid.laneCount[0] > 1) and (grid.laneCount[0] > 1) then begin
      for yi := 0 to grid.laneCount[1]-2 do for xi := 0 to grid.laneCount[0]-2 do begin
         // current spline area shape
        x1 := grid[0, xi];   x1_2 := x1*x1;  x1_3 := x1_2*x1;
        x2 := grid[0, xi+1]; x2_2 := x2*x2;  x2_3 := x2_2*x2;
        y1 := grid[1, yi];   y1_2 := y1*y1;  y1_3 := y1_2*y1;
        y2 := grid[1, yi+1]; y2_2 := y2*y2;  y2_3 := y2_2*y2;
        // right hand side
        b[0]  := sx[yi  ].splines[xi].calc(x1);
        b[1]  := sx[yi+1].splines[xi].calc(x1);
        b[2]  := sx[yi  ].splines[xi].calc(x2);
        b[3]  := sx[yi+1].splines[xi].calc(x2);
        b[4]  := sy[xi  ].splines[yi].derivative(y1);
        b[5]  := sy[xi  ].splines[yi].derivative(y2);
        b[6]  := sy[xi+1].splines[yi].derivative(y1);
        b[7]  := sy[xi+1].splines[yi].derivative(y2);
        b[8]  := sx[yi  ].splines[xi].derivative(x1);
        b[9]  := sx[yi+1].splines[xi].derivative(x1);
        b[10] := sx[yi  ].splines[xi].derivative(x2);
        b[11] := sx[yi+1].splines[xi].derivative(x2);
        b[12] := sd[yi  ].splines[xi].derivative(x1);
        b[13] := sd[yi+1].splines[xi].derivative(x1);
        b[14] := sd[yi  ].splines[xi].derivative(x2);
        b[15] := sd[yi+1].splines[xi].derivative(x2);
        // Matrix
        Mrow(0,   1, y1, y1_2,   y1_3,   x1, x1*y1, x1*y1_2,   x1*y1_3,   x1_2, x1_2*y1, x1_2*y1_2,   x1_2*y1_3,     x1_3,   x1_3*y1,   x1_3*y1_2,   x1_3*y1_3); // f(x1,y1)
        Mrow(1,   1, y2, y2_2,   y2_3,   x1, x1*y2, x1*y2_2,   x1*y2_3,   x1_2, x1_2*y2, x1_2*y2_2,   x1_2*y2_3,     x1_3,   x1_3*y2,   x1_3*y2_2,   x1_3*y2_3); // f(x1,y2)
        Mrow(2,   1, y1, y1_2,   y1_3,   x2, x2*y1, x2*y1_2,   x2*y1_3,   x2_2, x2_2*y1, x2_2*y1_2,   x2_2*y1_3,     x2_3,   x2_3*y1,   x2_3*y1_2,   x2_3*y1_3); // f(x2,y1)
        Mrow(3,   1, y2, y2_2,   y2_3,   x2, x2*y2, x2*y2_2,   x2*y2_3,   x2_2, x2_2*y2, x2_2*y2_2,   x2_2*y2_3,     x2_3,   x2_3*y2,   x2_3*y2_2,   x2_3*y2_3); // f(x2,y2)
        Mrow(4,   0,  1, 2*y1, 3*y1_2,    0,    x1, 2*x1*y1, 3*x1*y1_2,      0,    x1_2, 2*x1_2*y1, 3*x1_2*y1_2,        0,      x1_3,   2*x1_3*y1, 3*x1_3*y1_2); // fy(x1,y1)
        Mrow(5,   0,  1, 2*y2, 3*y2_2,    0,    x1, 2*x1*y2, 3*x1*y2_2,      0,    x1_2, 2*x1_2*y2, 3*x1_2*y2_2,        0,      x1_3,   2*x1_3*y2, 3*x1_3*y2_2); // fy(x1,y2)
        Mrow(6,   0,  1, 2*y1, 3*y1_2,    0,    x2, 2*x2*y1, 3*x2*y1_2,      0,    x2_2, 2*x2_2*y1, 3*x2_2*y1_2,        0,      x2_3,   2*x2_3*y1, 3*x2_3*y1_2); // fy(x2,y1)
        Mrow(7,   0,  1, 2*y2, 3*y2_2,    0,    x2, 2*x2*y2, 3*x2*y2_2,      0,    x2_2, 2*x2_2*y2, 3*x2_2*y2_2,        0,      x2_3,   2*x2_3*y2, 3*x2_3*y2_2); // fy(x2,y2)
        Mrow(8,   0,  0,    0,      0,    1,    y1,    y1_2,      y1_3,   2*x1, 2*x1*y1, 2*x1*y1_2,   2*x1*y1_3,   3*x1_2, 3*x1_2*y1, 3*x1_2*y1_2, 3*x1_2*y1_3); // fx(x1,y1)
        Mrow(9,   0,  0,    0,      0,    1,    y2,    y2_2,      y2_3,   2*x1, 2*x1*y2, 2*x1*y2_2,   2*x1*y2_3,   3*x1_2, 3*x1_2*y2, 3*x1_2*y2_2, 3*x1_2*y2_3); // fx(x1,y2)
        Mrow(10,  0,  0,    0,      0,    1,    y1,    y1_2,      y1_3,   2*x2, 2*x2*y1, 2*x2*y1_2,   2*x2*y1_3,   3*x2_2, 3*x2_2*y1, 3*x2_2*y1_2, 3*x2_2*y1_3); // fx(x2,y1)
        Mrow(11,  0,  0,    0,      0,    1,    y2,    y2_2,      y2_3,   2*x2, 2*x2*y2, 2*x2*y2_2,   2*x2*y2_3,   3*x2_2, 3*x2_2*y2, 3*x2_2*y2_2, 3*x2_2*y2_3); // fx(x2,y2)
        Mrow(12,  0,  0,    0,      0,    0,     1,    2*y1,    3*y1_2,      0,    2*x1,   4*x1*y1,   6*x1*y1_2,        0,    3*x1_2,   6*x1_2*y1, 9*x1_2*y1_2); // fxy(x1,y1)
        Mrow(13,  0,  0,    0,      0,    0,     1,    2*y2,    3*y2_2,      0,    2*x1,   4*x1*y2,   6*x1*y2_2,        0,    3*x1_2,   6*x1_2*y2, 9*x1_2*y2_2); // fxy(x1,y2)
        Mrow(14,  0,  0,    0,      0,    0,     1,    2*y1,    3*y1_2,      0,    2*x2,   4*x2*y1,   6*x2*y1_2,        0,    3*x2_2,   6*x2_2*y1, 9*x2_2*y1_2); // fxy(x2,y1)
        Mrow(15,  0,  0,    0,      0,    0,     1,    2*y2,    3*y2_2,      0,    2*x2,   4*x2*y2,   6*x2*y2_2,        0,    3*x2_2,   6*x2_2*y2, 9*x2_2*y2_2); // fxy(x2,y2)
        // solve linear system
        LinAlg.gauss(M, a, b, false);
        with (TSplPoly(spl[(grid.laneCount[0]-1)*yi + xi].asObject)) do begin
          co[0, 0] := a[0];
          co[0, 1] := a[1];
          co[0, 2] := a[2];
          co[0, 3] := a[3];
          co[1, 0] := a[4];
          co[1, 1] := a[5];
          co[1, 2] := a[6];
          co[1, 3] := a[7];
          co[2, 0] := a[8];
          co[2, 1] := a[9];
          co[2, 2] := a[10];
          co[2, 3] := a[11];
          co[3, 0] := a[12];
          co[3, 1] := a[13];
          co[3, 2] := a[14];
          co[3, 3] := a[15];
        end;
      end;
    end else if (grid.laneCount[0] = 1) and (grid.laneCount[0] = 1) then begin // special case: only one grid lane in x and y-direction
      p[0] := grid[0, 0];
      p[1] := grid[1, 0];
      with (TSplPoly(spl[0].asObject)) do begin
        co[0, 0] := orig.calc(p);
        co[0, 1] := 0;
        co[0, 2] := 0;
        co[0, 3] := 0;
        co[1, 0] := 0;
        co[1, 1] := 0;
        co[1, 2] := 0;
        co[1, 3] := 0;
        co[2, 0] := 0;
        co[2, 1] := 0;
        co[2, 2] := 0;
        co[2, 3] := 0;
        co[3, 0] := 0;
        co[3, 1] := 0;
        co[3, 2] := 0;
        co[3, 3] := 0;
      end;
    end else if (grid.laneCount[0] = 1) then begin // special case: only one grid lane in x-direction
      for yi := 0 to grid.laneCount[1]-2 do
        with (TSplPoly(spl[yi].asObject)) do begin
          co[0, 0] := sy[0].splines[yi].co[0];
          co[0, 1] := sy[0].splines[yi].co[1];
          co[0, 2] := sy[0].splines[yi].co[2];
          co[0, 3] := sy[0].splines[yi].co[3];
          co[1, 0] := 0;
          co[1, 1] := 0;
          co[1, 2] := 0;
          co[1, 3] := 0;
          co[2, 0] := 0;
          co[2, 1] := 0;
          co[2, 2] := 0;
          co[2, 3] := 0;
          co[3, 0] := 0;
          co[3, 1] := 0;
          co[3, 2] := 0;
          co[3, 3] := 0;
        end;
    end else if (grid.laneCount[1] = 1) then begin // special case: only one grid lane in y-direction
      for xi := 0 to grid.laneCount[0]-2 do
        with (TSplPoly(spl[xi].asObject)) do begin
          co[0, 0] := sx[0].splines[xi].co[0];
          co[0, 1] := 0;
          co[0, 2] := 0;
          co[0, 3] := 0;
          co[1, 0] := sx[0].splines[xi].co[1];
          co[1, 1] := 0;
          co[1, 2] := 0;
          co[1, 3] := 0;
          co[2, 0] := sx[0].splines[xi].co[2];
          co[2, 1] := 0;
          co[2, 2] := 0;
          co[2, 3] := 0;
          co[3, 0] := sx[0].splines[xi].co[3];
          co[3, 1] := 0;
          co[3, 2] := 0;
          co[3, 3] := 0;
        end;
    end else raise Exception.Create('Unknown error!');
  finally
    for xi := 0 to grid.laneCount[0]-1 do sy[xi].destroy;
    for yi := 0 to grid.laneCount[1]-1 do sx[yi].destroy;
    for yi := 0 to grid.laneCount[1]-1 do sd[yi].destroy;
  end;
end;

end.


